import java.util.Vector;

public class Display {
  
  private static final int CODE_DELAY = 0x00;
  private static final int CODE_CMD   = 0x02;
  private static final int CODE_DATA  = 0x03;

  private static final int NUM_LINES  =  2;
  private static final int LINE_WIDTH = 40;
  private static final int NUM_CHARS = NUM_LINES * LINE_WIDTH;
  private byte lines_[];

  private int pos_; // cursor position

  private static final int MODE_INC =  1;
  private static final int MODE_DEC = -1;
  private int moveMode_;

  private static final int MODE_CURSOR  = 4; 
  private static final int MODE_DISPLAY = 5;
  private int shiftMode_;

  private static final int BITS_8 = 8;
  private static final int BITS_4 = 4;
  private int interfaceBits_;

  private int brightnessPct_;

  private int ddAddr_;
  private int cgAddr_;

  private boolean displayOn_;
  private boolean cursorVisible_;
  private boolean blinkingEnabled_;
  
  public static interface ChangeListener {
    public void displayChanged();
    public void addNewChar(int code, VFDChar c);
  }
  private Vector changeListeners_;

  public Display() { 
    Util.check(NUM_CHARS > 0);
    changeListeners_ = new Vector();
    lines_ = new byte[LINE_WIDTH * NUM_LINES];
    clearDisplay();
    moveMode_ = MODE_INC;
    shiftMode_ = MODE_CURSOR;
    displayOn_ = true; 
    cursorVisible_ = false;
    blinkingEnabled_ = false;
    interfaceBits_ = BITS_8;
    brightnessPct_ = 100;
    cgAddr_ = ddAddr_ = 0; // XXX
  }

  private void clearDisplay() {
    for (int i = 0; i < LINE_WIDTH * NUM_LINES; i++)
      lines_[i] = ' ';
    pos_ = 0;
  }

  private void moveCursor(int amount) {
    pos_ += amount;
    if (pos_ < 0)
      pos_ = 0;
    if (pos_ > NUM_CHARS - 1)
      pos_ = NUM_CHARS - 1;
  }

  public int getX()                 { return pos_ % LINE_WIDTH; }
  public int getY()                 { return pos_ / NUM_LINES; }
  public int height()               { return NUM_LINES; }
  public int width()                { return LINE_WIDTH; }
  public boolean isOn() { 
    // return displayOn_; 

    // Display is always ``on'', e.g. to show clock.  Maybe the real
    // device uses the ``display command'' to power down some of the
    // hardware?
    return true;
  }
  public boolean cursorVisible()    { return cursorVisible_; }
  public boolean blinkingEnabled()  { return blinkingEnabled_; }
  public int brightness()           { return brightnessPct_; } // percentage 0--100

  public String toString() {
    StringBuffer s = new StringBuffer(NUM_CHARS + NUM_LINES);
    for (int i = 0; i < NUM_CHARS; i++) {
      if (i > 0 && (i % LINE_WIDTH) == 0)
	s.append('\n');
      s.append(translateChar(lines_[i]));
    }
    s.append('\n');
    return s.toString();
  }

  public byte getByteAt(int x, int y) {
    Util.check(x >= 0 && x < LINE_WIDTH);
    Util.check(y >= 0 && y < NUM_LINES);
    int pos = x + y*LINE_WIDTH;
    return lines_[pos];
  }

  public char translateChar(byte b) {
    int c = 0xff & b;
    if (c == 0x8c) return '\\';
    else if (c == 0x8e) return '~';
    else if (c == 0xf0) return 'p';
    else if (c == 0xf1) return 'q';
    else if (c == 0xf9) return 'y';
    else return (char) c;
  }

  public void handleCodes(byte codes[], int offset, int n) {
    Util.check(offset >= 0);
    Util.check(n >= 0);
    Util.check(offset + n <= codes.length);
    Util.check((n & 1) == 0);
    
    int numCodes = n / 2;
    for (int i = 0; i < numCodes; i++) {
      int c1 = 0xff & codes[i*2 + offset];
      int c2 = 0xff & codes[i*2 + offset + 1];

      if (c1 == CODE_DELAY) 
	; // punt
      else if (c1 == CODE_CMD) 
	i += handleCommand(c2, codes, offset + 2*i, n - 2*i);
      else if (c1 == CODE_DATA)
	handleData(c2);    
      else 
	Util.dbg("Display.handleCode(): Unknown code c1=0x" +
		 Integer.toHexString(c1) + ", c2=0x" + Integer.toHexString(c2));
    }
  }

  public void addChangeListener(ChangeListener l) {
    changeListeners_.add(l);
  }

  // called when display contents change
  public void displayChanged() {
    for (int i = 0; i < changeListeners_.size(); i++)
      ((ChangeListener) changeListeners_.get(i)).displayChanged();
  }

  // called when new char defined
  public void addNewChar(int code, VFDChar c) {
    for (int i = 0; i < changeListeners_.size(); i++)
      ((ChangeListener) changeListeners_.get(i)).addNewChar(code, c);
  }

  private void handleData(int data) {
    byte c = (byte) (data & 0xff);
    lines_[pos_] = c;
    if (moveMode_ == MODE_INC)
      moveCursor(1);
    else if (moveMode_ == MODE_DEC)
      moveCursor(-1);
    else
      Util.check(0);
    displayChanged();
  }

  private static final int CMD_CLEAR    = 0;  // Clear display 
  private static final int CMD_HOME     = 1;  // Goto home position
  private static final int CMD_MODE     = 2;  // Set cursor & shift mode
  private static final int CMD_DISPLAY  = 3;  // Set display mode
  private static final int CMD_SHIFT    = 4;  // Shift display/cursor
  private static final int CMD_FUNC_SET = 5;  // Select bus width, change brightness
  private static final int CMD_CG_ADDR  = 6;
  private static final int CMD_DD_ADDR  = 7;

  private boolean isCmd(int cmd, int possibleCommand) {
    return (cmd >> possibleCommand) == 1;
  }

  private int handleCommand(int cmd, byte codes[], int offset, int n) {
    int extraConsumed = 0;

    if (isCmd(cmd, CMD_CLEAR))
      clearDisplay();
    else if (isCmd(cmd, CMD_HOME))
      pos_ = 0;
    else if (isCmd(cmd, CMD_MODE))
      handleModeCommand(cmd);
    else if (isCmd(cmd, CMD_DISPLAY))
      handleDisplayCommand(cmd);
    else if (isCmd(cmd, CMD_SHIFT))
      handleShiftCommand(cmd);
    else if (isCmd(cmd, CMD_FUNC_SET)) 
      extraConsumed = handleFuncSetCommand(cmd, codes, offset, n);
    else if (isCmd(cmd, CMD_CG_ADDR))
      extraConsumed = handleCGAddrCommand(cmd, codes, offset, n);
    else if (isCmd(cmd, CMD_DD_ADDR))
      handleDDAddrCommand(cmd);
    else if (cmd != 0)
      Util.dbg("unknown cmd: 0x" + Integer.toHexString(cmd)); // punt
    return extraConsumed;
  }

  private void handleModeCommand(int cmd) {
    moveMode_ = ((cmd & 2) != 0) ? MODE_INC : MODE_DEC;
    shiftMode_ = ((cmd & 1) != 0) ? MODE_DISPLAY : MODE_CURSOR;
  }

  private void handleDisplayCommand(int cmd) {
    boolean newDisplayOn = (cmd & 4) != 0;
    boolean newCursorVisible = (cmd & 2) != 0;
    boolean newBlinkingEnabled = (cmd & 1) != 0;

    boolean changed = newDisplayOn != displayOn_ 
      || newCursorVisible != cursorVisible_ 
      || newBlinkingEnabled != blinkingEnabled_;

    displayOn_ = newDisplayOn;
    cursorVisible_ = newCursorVisible;
    blinkingEnabled_ = newBlinkingEnabled;
    
    if (changed)
      displayChanged();
  }

  private void handleShiftCommand(int cmd) {
    if ((cmd & 8) != 0) {
      // punt, this is ``display shift'', whatever that is
      Util.dbg("Display.handleShiftCommand: Display shift not implemented!");
    }
    else if ((cmd & 4) != 0) {
      // shift right
      moveCursor(1);
      if (isOn() && (cursorVisible() || blinkingEnabled()))
	displayChanged();
    }
    else {
      // shift left 
      moveCursor(-1);
      if (isOn() && (cursorVisible() || blinkingEnabled()))
	displayChanged();
    }    
  }

  private int handleFuncSetCommand(int cmd, byte codes[], int offset, int n) {
    interfaceBits_ = ((cmd & 16) != 0) ? BITS_8 : BITS_4;
    if ((cmd & 0xf) != 0)
      return 0;

    // otherwise look for brightness command following
    if (n < 4) {
      Util.dbg("Display.handleFuncSetCommand: missing brightness data");
      return 0;
    }
    
    int c0 = codes[offset + 2] & 0xff;
    int c1 = codes[offset + 3] & 0xff;
    
    if (c0 == 0x3) {
      brightnessPct_ = 100 - 25 * (c1 & 0x3);
      displayChanged();
      return 1;	
    }
    else {
      Util.dbg("Display.handleFuncSetCommand: unexpected brightness data format");
      return 0;
    }
  }

  private void handleDDAddrCommand(int cmd) {
    int mask = 0xff >> 1;
    ddAddr_ = mask & cmd;
  }

  private int handleCGAddrCommand(int cmd, byte codes[], int offset, int n) {
    if (n < 18) {
      Util.dbg("Display.handleCGAddrCommand: missing font data");
      return 0;
    }
    
    int mask = 0xff >> 2;
    cgAddr_ = (mask & cmd) / 8;
    VFDChar c = new VFDChar();
    for (int i = 0; i < 7; i++) {
      int cmd2 = 0xff & codes[offset + 2 + 2*i];
      if (cmd2 != CODE_DATA)
	Util.dbg("Display.handleCGAddrCommand: confusing character data");
      int line = 0xff & codes[offset + 2 + 2*i + 1];
      for (int j = 0; j < 5; j++) {
	boolean bit = (1 & (line >> (5 - j - 1))) != 0;
	c.setDot(j, i, bit);
      }
    }
    if ((0x2 & codes[offset + 2 + 14 + 1]) != 0)
      c.setUnderline(true);
    else
      c.setUnderline(false);
    
    addNewChar(cgAddr_, c);
    return 8;
  }
}

